import tensorflow as tf
import os
import numpy as np

def double_conv(in_channels, out_channels):
    return tf.keras.Sequential(layers=(tf.keras.layers.Conv2D(out_channels, kernel_size=3, padding='SAME'),
                                       tf.keras.layers.BatchNormalization(),
                                       tf.keras.layers.ReLU(),
                                       tf.keras.layers.Conv2D(out_channels, kernel_size=3, padding='SAME'),
                                       tf.keras.layers.BatchNormalization(),
                                       tf.keras.layers.ReLU()
                                       ))

class Down(tf.keras.layers.Layer):
    def __init__(self, in_channels, out_channels):
        super(Down, self).__init__()
        self.conv = double_conv(in_channels, out_channels)
        self.pool = tf.keras.layers.MaxPool2D(2)

    def call(self, x):
        x = self.pool(x)
        return self.conv(x)


class Up(tf.keras.layers.Layer):
    def __init__(self, in_channels, out_channels):
        super(Up, self).__init__()
        self.conv = double_conv(in_channels, out_channels)
        self.upsample = tf.keras.layers.UpSampling2D(2, interpolation='bilinear')

    def call(self, x1, x2):
        x1 = self.upsample(x1)
        diffY = x2.shape[1] - x1.shape[1]
        diffX = x2.shape[2] - x1.shape[2]
        paddings = tf.constant([[0, 0], [diffY, 0], [diffX, 0], [0, 0]])
        x1 = tf.pad(x1, paddings)
        x = tf.concat((x2, x1), axis=-1)
        return self.conv(x)


class UNet(tf.keras.Model):
    def __init__(self, device = None):
        super(UNet, self).__init__()
        self.inc = double_conv(3, 4)  
        self.down1 = Down(4, 8)
        self.down2 = Down(8, 16)
        self.down3 = Down(16, 32)
        factor = 2
        self.down4 = Down(32, 64 // factor)
        self.up1 = Up(64, 32 // factor)
        self.up2 = Up(32, 16 // factor)
        self.up3 = Up(16, 8 // factor)
        self.up4 = Up(8, 4)
        self.outc = tf.keras.layers.Conv2D(filters=1, kernel_size=1)

    def call(self, x, training=False):
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        logits = self.outc(x)
        return tf.sigmoid(logits)

    def loadModel(self, path):
        if os.path.isdir(path):
            latest = tf.train.latest_checkpoint(path)
            self.load_weights(latest)

    def loadImage(self, path):
        tmp = tf.keras.preprocessing.image.load_img(path)
        x = tf.keras.preprocessing.image.img_to_array(tmp) / 255
        x = tf.reshape(x, (1, x.shape[0], x.shape[1], x.shape[2]))
        test_images_tensor = tf.data.Dataset.from_tensors(x)
        return test_images_tensor


import os
import numpy as np

import tensorflow as tf
from PIL import Image
from unet import UNet

from utils import load_data, plot_acc_loss

os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

gpu_available = tf.test.is_gpu_available(cuda_only=True)

if gpu_available:
    print('GPU found')
    device = 'gpu'
else:
    print("No GPU found")
    device = 'cpu'


PATH = os.getcwd()
BATCH_SIZE = 16
EPOCHS_NUMBER = 2000
LOSS = 'bce'
SAVING_NAME = 'model{}_e{}_b{}'.format(LOSS, EPOCHS_NUMBER, BATCH_SIZE)

print("Start load train images.")
train_images = load_data('train_images') 
train_masks = load_data('train_masks')
print("Start load test images.")
test_images = load_data('test_images') 
test_masks = load_data('test_masks') 
print("End load.")

model = UNet()

trained = False
model_path = os.path.join(os.getcwd(), "src", "Unet", "models")
# проверку переделать
if len(os.listdir(model_path)) != 0:
    latest = tf.train.latest_checkpoint(model_path)
    model.load_weights(latest)
    trained = True
 
if not trained:
    
    model.compile(optimizer='adam',
                loss=[LOSS],
                metrics=['bce', 'accuracy', tf.keras.metrics.Precision(), tf.keras.metrics.Recall()])

    # Train the model
    print("Start train!")

    dataset = tf.data.Dataset.from_tensors((train_images, train_masks))
    test_dataset = tf.data.Dataset.from_tensors((test_images, test_masks))

    print(dataset)

    fit = model.fit(dataset, batch_size=BATCH_SIZE, epochs=EPOCHS_NUMBER, shuffle=True, validation_data=test_dataset)
    
    test_loss, _, test_acc, _, _   = model.evaluate(test_dataset)

    # save model
    
    model.save_weights('src/Unet/models/{}model'.format(SAVING_NAME))

    print("Train complete. Model saved.")

test_images_tensor = tf.data.Dataset.from_tensors(test_images)
# # Predict and save masks for the test set
predicted_masks = model.predict(test_images_tensor, batch_size=BATCH_SIZE)

for i in range(10):
    pred_mask = np.squeeze(predicted_masks[i])
    true_mask = test_masks[i]
    tmp = np.ceil(pred_mask * 255.).astype(np.uint8)
    im = Image.fromarray(tmp)
    im.save('src/Unet/results/test{}{}.{}.png'.format(EPOCHS_NUMBER, BATCH_SIZE, i))

# results_file = open('models/{}test_loss{}test_accuracy.txt'.format(test_loss, test_acc), 'w')
# results_file.write(SAVING_NAME)
# results_file.close()

# plot_acc_loss(fit.history['accuracy'], fit.history['loss'], SAVING_NAME)



import os
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt

from tensorflow import keras

PATH = os.getcwd()

def load_data(folder_name):
    folder_path = os.path.join(os.path.dirname(__file__), "data", folder_name)
    x_data = []
    for sample in sorted(os.listdir(folder_path)):
        tmp = tf.keras.preprocessing.image.load_img(os.path.join(folder_path, sample))
        tmp = tf.image.resize(tmp, [800, 600])
        x = tf.keras.preprocessing.image.img_to_array(tmp) / 255
        
        if(folder_name == "train_masks" or folder_name == "test_masks"):
            converted = tf.image.rgb_to_grayscale(x)
            x_data.append(converted)
        else:
           x_data.append(x) 
    return x_data

def plot_acc_loss(accuracy_values, loss_values, saving_name):
    # plot training ACCURACY VALUES
    fig = plt.figure()
    gs = fig.add_gridspec(ncols = 1,nrows = 2)
    plt.subplot(gs[0])
    plt.plot(accuracy_values)
    plt.ylabel('Accuracy')
    plt.legend(['Train data'], loc = 'upper left')

    # plot training LOSS VALUES
    plt.subplot(gs[1])
    plt.plot(loss_values)
    plt.ylabel('Loss')
    plt.xlabel('Epoch')
    plt.legend(['Train data'], loc = 'upper left')
    plt.tight_layout()
    fig.savefig(os.path.join(PATH, 'results','') + '{}.png'.format(saving_name), dpi = fig.dpi)